import time
import nemu
import weakref
import ipaddress
from subprocess import PIPE
from pathlib2 import Path

fix_file =  Path(__file__).parent.resolve() / "fixnemu.py"
execfile(str(fix_file))
IPTABLES = 'iptables'

class Node(nemu.Node):
    """simple nemu.Node used for registry and nodes"""
    def _add_interface(self, iface):
        self.iface = iface
        iface.__dict__['node'] = weakref.proxy(self)
        return super(Node, self)._add_interface(iface)

    @property
    def ip(self):
        if hasattr(self, "_ip"):
            return str(self._ip)

        # return 1 ipv4 address of the one interface, reverse mode
        for iface in self.get_interfaces()[::-1]:
            for addr in iface.get_addresses():
                addr = addr['address']
                if '.' in addr:
                    #TODO different type problem?
                    self._ip = ipaddress.ip_address(addr)
                    return addr

    def connect_network(self, internet):
        """ create a p2p interface with internet, and asgin ipaddress for
        interface

        internet: Internet object
        ret: 2 interface
        """
        self.if_i, it_if = nemu.P2PInterface.create_pair(self, internet)

        ip = internet.assign_ip(8)
        self._ip = ip + 2
        self.if_i.add_v4_address(str(self._ip), prefix_len=24)
        it_if.add_v4_address(str(ip+1), prefix_len=24)
        internet.iface[self] = it_if
        self.if_i.up = it_if.up = True

        self.add_route(prefix=str(internet.net[0]) ,prefix_len=internet.net.prefixlen,
                       nexthop=str(ip+1))

        return self.if_i, it_if


    def connect_switch(self, switch, ip):
        self.if_s = if_s = nemu.NodeInterface(self)
        switch.connect(if_s)
        if_s.up = True
        if_s.add_v4_address(ip, prefix_len=24)
        return if_s


class Network(nemu.Node):

    def __init__(self, net):
        """net: ipv4 network like 10.0.0.0/8"""
        super(Network, self).__init__()
        self.net = ipaddress.ip_network(u"{}".format(net))
        # start at 10.1.0.0
        self.next_net = self.net[1<<16]
        self.iface = {}

    def assign_ip(self, length):
        ret = self.next_net
        self.next_net += 1 << length
        return ret

class NetManager(object):
    """contain all the nemu object created, so they can live more time"""
    def __init__(self):
        self.object = []
        self.registrys = {}

def network_demo():
    registry = Node()
    internet = Network("10.0.0.0/8")

    m0 = Node()
    m1 = Node()
    m2 = Node()
    m3 = Node()
    m4 = Node()
    m5 = Node()
    # gateway
    g1 = Node()

    switch0 = nemu.Switch()
    switch1 = nemu.Switch()

    nm = NetManager()
    nm.object = [registry, internet, m0, m1, m2, m3, m4, m5, g1, switch0, switch1]
    nm.registrys = {registry: [m0, m1, m2, m3, m4, m5]}

    
    for node in [registry, internet, m0, m2, m3, m4, m5]:
        print "pid: {}".format(node.pid)

    g0_if_0, _ = g1.connect_network(internet)
    m0.connect_network(internet)
    registry.connect_network(internet)

    g1.Popen((IPTABLES, '-t', 'nat', '-A', 'POSTROUTING', '-o', g0_if_0.name, '-j', 'MASQUERADE')).wait()
    g1.Popen((IPTABLES, '-t', 'nat', '-N', 'MINIUPNPD')).wait()
    g1.Popen((IPTABLES, '-t', 'nat', '-A', 'PREROUTING', '-i', g0_if_0.name, '-j', 'MINIUPNPD')).wait()
    g1.Popen((IPTABLES, '-N', 'MINIUPNPD')).wait()
    # Enable forwarding for communication between registry and registry2
    internet.Popen(('sysctl', '-q', 'net.ipv6.conf.all.forwarding=1')).wait()

    it_if_2 = nemu.NodeInterface(internet)
    switch0.connect(it_if_2)
    it_if_2.add_v4_address("10.1.9.1", prefix_len=24)
    it_if_2.up = True
    nm.object.append(it_if_2)


    ip = ipaddress.ip_address(u"10.1.9.1")
    for node in [m1, m2]:
        ip += 1
        node.connect_switch(switch0, str(ip))


    ip = ipaddress.ip_address(u"10.1.10.0")
    for node in [g1, m3, m4, m5]:
        ip += 1
        node.connect_switch(switch1, str(ip))


    switch0.up = True
    switch1.up = True

    for node in [m1, m2]:
        node.add_route(nexthop="10.1.9.1")
    
    for node in [m3, m4, m5]:
        node.add_route(nexthop="10.1.10.1")



    for node in [m0, m1, m2, m3, m4, m5]:
        app0 = node.Popen(["ping", "-c", "1", registry.ip], stdout=PIPE)
        ret = app0.wait()
        assert ret == 0, "network construct failed"

    time.sleep(1000000)
    return nm


def network_direct():
    """one server and one client connect direct"""
    registry = Node()
    m0 = Node()
    nm = NetManager()
    nm.registrys = {registry: [m0]}

    re_if_0, m_if_0 = nemu.P2PInterface.create_pair(registry, m0)
    
    registry._ip = u"10.1.2.1"
    re_if_0.add_v4_address(u"10.1.2.1", prefix_len=24)
    
    m_if_0.add_v4_address(u"10.1.2.2", prefix_len=24)
    re_if_0.up = m_if_0.up = True

    for node in [m0]:
        app0 = node.Popen(["ping", "-c", "1", "10.1.2.1"], stdout=PIPE)
        ret = app0.wait()
        assert ret == 0, "network construct failed"
    return nm
        
if __name__ == "__main__":
    network_demo()
